

//------------------------------------------------------------------
#define  FT201WR     0x22 << 1
#define FT201RD     FT201WR|1
#define FT201REPEAT   3


//======================================================================//
#define  TWISCK  400     // CLK w kHz

#include <avr/io.h>                  // IO
#include <stdio.h>                  //
#include <stdlib.h>                 //
//#include <ctype.h>
#include <avr/sfr_defs.h>             // SFR, _BV(bit)


unsigned char TWI_Status = -1;


void TWI_Master_Initialise();
void TWI_Master_Off();
unsigned char TWI_Action( unsigned char command );
unsigned char TWI_Start();
unsigned char TWI_Stop();
#define  TWI_Exit  TWI_Stop
unsigned char TWI_Write( unsigned char d );
unsigned char TWI_Read( unsigned char ack );


//======================================================================//
void TWI_Master_Initialise()
{
  TWI_Master_Off();

  //SCL frequency CPU Clock frequency / ( 16 + 2(TWBR) + 4 � TWPS ) TWPS = PRESKALER
  //F_CPU/(16+2*TWBR+4*(TWSR&3))
  // TWBR min 10
  TWBR = ((F_CPU / 1000 / TWISCK) - 16) / 2; // TWISCK - CLK w kHz
#define TWBR_100KHZ   ((F_CPU/1000/100)-16)/2;
#define TWBR_400KHZ   ((F_CPU/1000/400)-16)/2;
#if (((F_CPU/1000/TWISCK)-16)/2) < 10
  //    #error "Podobno sieje bledy"
#warning "Podobno sieje bledy"
  //  if ( TWBR < 10 ) TWBR=10;
#endif

  // TWSR = TWI_TWPS;   // Not used. Driver presumes prescaler to be 00.
  //  TWDR = 0xFF;      // Default content = SDA released.
  TWCR = (1 << TWEN); // Enable TWI-interface and release TWI pins.
}


//----------------------------------------------------------------------//
void TWI_Master_Off()
{
  /*
    TWCR = 0;               // Disable TWI-interface and release TWI pins.
    #if defined(__AVR_ATmega164P__) || defined(__AVR_ATmega324P__) || defined(__AVR_ATmega1284P__)
    #define DDR_TWI   DDRC
    #define PORT_TWI  PORTC
    #define SCL_TWI   PC0
    #define SDA_TWI   PC1
    #elif defined(__AVR_ATmega128P__) || defined(__AVR_ATmega1281__)  || defined(__AVR_ATmega1281P__) || defined(__AVR_ATmega2561P__)
    #define DDR_TWI   DDRD
    #define PORT_TWI  PORTD
    #define SCL_TWI   PD0
    #define SDA_TWI   PD1
    #else
    #error "Nieobsluzony CPU"
    #endif
    //todo: mo�na by poni�sze pomin�c gdy SDA jest w stanie wysokiem
    PORT_TWI |= _BV(SDA_TWI) | _BV(SCL_TWI);  // Stan wysoki IIC
    DDR_TWI |= _BV(SDA_TWI) | _BV(SCL_TWI);   // Port IIC wyj�ciem
    for(unsigned char x=0; x<9; x++)            // 9 impuls�w na SCL w razie gdy SLAVE zablokowany podczas nadawania gdy SDA w niski
    {
    PORT_TWI &= ~_BV(SCL_TWI);        // SCL niski
    _delay_us(10);              // CLK 100kHz
    PORT_TWI |= _BV(SCL_TWI);       // SCL niski
    _delay_us(10);              // CLK 100kHz
    }
    PORT_TWI |= _BV(SDA_TWI) | _BV(SCL_TWI);  // Stan wysoki IIC
    DDR_TWI &= ~(_BV(SDA_TWI) | _BV(SCL_TWI));  // Port IIC wej�ciem
  */
}


//----------------------------------------------------------------------//
unsigned char TWI_Action( unsigned char command )
{
  unsigned char overtime = 100;


  //write command to TWCR and make sure TWINT is set
  TWCR = (command | (1 << TWINT) | (1 << TWEN) );
  //now wait for TWINT to be set again (when the operation is completed)
  while ( !(TWCR & (1 << TWINT)) )
  {
    delayMicroseconds( 10 ); //_delay_us(10);
    overtime--;
    if ( !overtime ) break;
  }
  /*
    There are two status codes that do not correspond to a defined TWI state, see Table
    120.
    Status 0xF8 indicates that no relevant information is available because the TWINT Flag
    is not set. This occurs between other states, and when the TWI is not involved in a serial
    transfer.
    Status 0x00 indicates that a bus error has occurred during a 2-wire Serial Bus transfer.
    A bus error occurs when a START or STOP condition occurs at an illegal position in the
    format frame. Examples of such illegal positions are during the serial transfer of an
    address unsigned char, a data unsigned char, or an acknowledge bit. When a bus error occurs, TWINT is
    set. To recover from a bus error, the TWSTO Flag must set and TWINT must be cleared
    by writing a logic one to it. This causes the TWI to enter the not addressed Slave mode
    and to clear the TWSTO Flag (no other bits in TWCR are affected). The SDA and SCL
    lines are released, and no STOP condition is transmitted.
  */
  /*
    Istniej� dwa kody stanu, kt�re nie odpowiadaj� okre�lonym stanie TWI, patrz tabela 120.

    Stan 0xF8 wskazuje, �e �adne istotne informacje s� dost�pne poniewa� TWINT Flag
    nie jest ustawiona. Dzieje si� tak gdy TWI nie jest zaangarzowany w transfer.
    Je�li status F8 pojawia si� po wys�aniu warunku statt, mo�e to oznacza� uszkodzenie magistrali IIC (zwarcie lub brak rezystor�w podci�gaj�cych).

    Stan 0x00 oznacza, �e wyst�pi� b��d magistrali IIC.
    B��d wyst�puje, gdy START lub STOP pojawi� si� w nieodpowiednim miejscu najcz�ciej podczas nielegalnego przej�cia magistrali.
    Gdy wyst�pi b��d magistrali (w trybie Slave), nale�y wyserowa� TWSTO i ustawiC TWINT.
    TWI nie wygeneruje nic na magistrali ale zwolni linie.
  */
  TWI_Status = TWSR & 0b11111100;
  return ( TWI_Status );
}


//----------------------------------------------------------------------//
unsigned char TWI_Stop()
{
  unsigned char overtime = 100;


  TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);
  // Generuj�c stop czekamy na bit TWSTO a nie WTINT
  while ( (TWCR & (1 << TWSTO)) )
  {
    delayMicroseconds( 10 ); //_delay_us(10);
    overtime--;
    if ( !overtime ) break;
  }


  return ( TWSR );
}


//----------------------------------------------------------------------//
unsigned char TWI_StatusTransceiver_Busy()
{
  return ( TWCR & (1 << TWIE) );  // IF TWI Interrupt is enabled then the Transceiver is busy
}


//----------------------------------------------------------------------//
unsigned char TWI_Start()
{
  return ( TWI_Action( 1 << TWSTA ) );
}


//----------------------------------------------------------------------//
unsigned char TWI_Read( unsigned char ack )
{
  if ( ack ) TWI_Action( 1 << TWEA );
  else TWI_Action( 0 );

  TWI_Status = TWSR;
  return ( TWDR );
}


//----------------------------------------------------------------------//
unsigned char TWI_Write( unsigned char d )
{
  TWDR = d;
  return ( TWI_Action( 0 ) );
}


//======================================================================//
word ReadDataFT201()
{
  byte st, dana = 0;


  //----- START -----//
  st = TWI_Start(); if ( st != 8 && st != 0x10 ) return ( 0 );
  st = TWI_Write( FT201RD );
  //    if ( st == 0x48 ){ TWI_Exit(); return( 0 ); }  // Brak ACK oznacza brak znaku do odbioru
  if ( st != 0x40 ) {
    TWI_Exit();
    return ( 0 );
  }
  // Odczytaj ostatni bajt
  dana = TWI_Read( false ); if ( TWI_Status != 0x58  ) {
    TWI_Exit();
    return ( 0 );
  }
  //----- STOP -----//
  st = TWI_Stop();


  return ( dana );
}


//-----------------------------------------------------------------
byte PutCharFT201( byte znak )
{
  byte st;


  //  if( StUsb != 3 ) return( false );

  //----- START -----//
  st = TWI_Start(); if ( st != 8 && st != 0x10 ) {
    TWI_Exit();
    return ( false );
  }
  //str.27  "AN_255_USB to I2C Example using the FT232H and FT201X devices"
  st = TWI_Write( FT201WR );  if ( st != 0x18 ) {
    TWI_Exit();  // Brak ACK to bufor nadawczy pełny
    return ( false );
  }
  st = TWI_Write( znak ); if ( st != 0x28 ) {
    TWI_Exit();
    return ( false );
  }
  //----- STOP -----//
  st = TWI_Stop();


  return ( true );
}


//-----------------------------------------------------------------
void PrintStringFT201( char *text )
{
  char znak;


  //  if( StUsb != 3 ) return;

  while (*text)         // Wyswietl tekst z RAMu
  {
    znak = (*text);
    PutCharFT201( znak );
    text++;
  }
}


//----------------------------------------------------------------------//
byte AvailableFT201()
{
  byte st, ava, repeat = FT201REPEAT;


  do
  {
    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // General adres
    st = TWI_Write( 0 );  if ( st != 0x18 ) {
      TWI_Exit();
      continue;
    }
    //str.25  "AN_255_USB to I2C Example using the FT232H and FT201X devices"
    //DataAvailable command
    st = TWI_Write( 0x0C ); if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }
    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) return ( 0 );
    st = TWI_Write( FT201RD );  if ( st != 0x40 ) {
      TWI_Exit();
      continue;
    }
    // Odczytaj ostatni bajt
    ava = TWI_Read( false ); if ( TWI_Status != 0x58  ) {
      TWI_Exit();
      continue;
    }
    //----- STOP -----//
    st = TWI_Stop();
    return ( ava );
  } while ( repeat-- );


  return 0;
}


//-------------------- czytaj status --------------------------------
byte StFT201;
byte ReadStFT201()
{
  byte st, repeat = FT201REPEAT;


  StFT201 = -1;

  do
  {
    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // General adres
    st = TWI_Write( 0 );  if ( st != 0x18 ) {
      TWI_Exit();
      continue;
    }
    //str.22  "AN_255_USB to I2C Example using the FT232H and FT201X devices"
    //cmd status
    st = TWI_Write( 0x16 ); if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }
    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;

    st = TWI_Write( FT201RD );  if ( st != 0x40 ) {
      TWI_Exit();
      continue;
    }
    // Odczytaj ostatni bajt
    StFT201 = TWI_Read( false ); if ( TWI_Status != 0x58  ) {
      TWI_Exit();
      continue;
    }
    //0x00 Suspended
    //0x01 Default
    //0x02 Addressed
    //0x03 Configured
    //----- STOP -----//
    st = TWI_Stop();
  } while ( repeat-- );


  return ( StFT201 );
}


//----------------------- czytaj ID ----------------------------
byte IdFT201[3];
byte ReadIdFT201()
{
  byte st;


  IdFT201[0] = IdFT201[1] = IdFT201[2] = -1;

  //----- START -----//
  st = TWI_Start(); if ( st != 8 && st != 0x10 ) return;
  //str.22  "AN_255_USB to I2C Example using the FT232H and FT201X devices"
  //cmd read ID
  st = TWI_Write( 0xF8/*0x7C*/ ); if ( st != 0x18 ) {
    TWI_Exit();
    return;
  }
  st = TWI_Write( FT201RD );  if ( st != 0x28 ) {
    TWI_Exit();
    return;
  }

  //----- START -----//
  st = TWI_Start(); if ( st != 8 && st != 0x10 ) return ( st );
  //ID command 0x7C
  st = TWI_Write( 0xF9/*0x7c*/ ); if ( st != 0x40 ) {
    TWI_Exit();
    return;
  }
  IdFT201[0] = TWI_Read( true ); if ( TWI_Status != 0x50  ) {
    TWI_Exit();
    return;
  }
  IdFT201[1] = TWI_Read( true ); if ( TWI_Status != 0x50  ) {
    TWI_Exit();
    return;
  }
  // Odczytaj ostatni bajt
  IdFT201[2] = TWI_Read( false ); if ( TWI_Status != 0x58  ) {
    TWI_Exit();
    return;
  }
  //ID[0] - pewnie vid pid jeśli różny od domyślnego
  //ID[1] -
  //ID[2] - wartośc ustawiana w TF_PROG
  //----- STOP -----//
  st = TWI_Stop();
}


//------------------------------------------------------------------
void FlushFT201()
{
  byte st, repeat = FT201REPEAT;


  do
  {
    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // General adres
    st = TWI_Write( 0 );  if ( st != 0x18 ) {
      TWI_Exit();
      continue;
    }
    // $20 - nie było ACK
    // $38 - przegrany arbritarz
    // CMD flush
    st = TWI_Write( 0x0E ); if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }
    //----- STOP -----//
    st = TWI_Stop();
  } while ( repeat-- );
}


//------------------------------------------------------------------
word ReadMtpFT201( word adres, word len, byte *buf )
{
  word x, d;


  for (x = 0; x < len; x++)
  {
    *buf++ = d = RdMtpFT201(adres + x);
    if ( d == 0 ) return ( x );
  }

  return ( x );
}


//------------------------------------------------------------------
word RdMtpFT201( word adres )
{
  byte st, dat, repeat = FT201REPEAT;


  do
  {
    if ( !SetMtpAdrFT201( adres )) continue;

    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // General adres
    st = TWI_Write( 0 ); if ( st != 0x18 ) {
      TWI_Exit();
      continue;
    }
    // CMD RD mem
    st = TWI_Write( 0x14 ); if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }

    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // Slave adres RD
    st = TWI_Write( FT201RD );  if ( st != 0x40 ) {
      TWI_Exit();
      continue;
    }

    // Odczytaj ostatni bajt
    dat = TWI_Read( false ); if ( TWI_Status != 0x58  ) {
      TWI_Exit();
      continue;
    }

    return ( dat );
  } while ( repeat-- );
  //----- STOP -----//
  st = TWI_Stop();

  return ( 0 );
}


//------------------------------------------------------------------
byte SetMtpAdrFT201( word adres )
{
  byte st, repeat = FT201REPEAT;


  do
  {
    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // General adres
    st = TWI_Write( 0 );  if ( st != 0x18 ) {
      TWI_Exit();
      continue;
    }
    // $20 - nie było ACK
    // $38 - przegrany arbritarz
    // CMD adres MTP
    st = TWI_Write( 0x10 ); if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }

    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // Slave adres WR
    st = TWI_Write( FT201WR );  if ( st != 0x18 ) {
      TWI_Exit();
      continue;
    }
    // MSB
    st = TWI_Write( adres >> 8 ); if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }
    // LSB
    st = TWI_Write( adres );  if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }

    //----- STOP -----//
    st = TWI_Stop();
    return true;
  } while ( repeat-- );


  return ( false );
}


//------------------------------------------------------------------
byte WriteMtpFT201( word adres, byte data )
{
  byte st, repeat = FT201REPEAT;


  do
  {
    if ( !SetMtpAdrFT201( adres ) ) continue;
    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // General adres
    st = TWI_Write( 0 ); if ( st != 0x18 ) {
      TWI_Exit();
      continue;
    }
    // CMD WR mem
    st = TWI_Write( 0x12 ); if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }

    //----- START -----//
    st = TWI_Start(); if ( st != 8 && st != 0x10 ) continue;
    // Slave adres WR
    st = TWI_Write( FT201WR );  if ( st != 0x18 ) {
      TWI_Exit();
      continue;
    }

    st = TWI_Write( data ); if ( st != 0x28 ) {
      TWI_Exit();
      continue;
    }

    //----- STOP -----//
    //  st = TWI_Stop();
    return ( true );
  } while ( repeat-- );


  return ( false );
}


//======================================================================//
void setup() {
  TWI_Master_Initialise();

  Serial.begin( 115200 );
  Serial.print( "Start programu\n\r" );


  char txt[100];
  ReadIdFT201();
  sprintf( txt, "Status FT201 %d ID:%02x%02x%02x\n\r",  ReadStFT201(), IdFT201[0], IdFT201[1], IdFT201[2] );
  Serial.print( txt );


  uint8_t static cntFT201;
  ReadMtpFT201( 0x18 << 1, 1, &cntFT201 );
  cntFT201++;
  WriteMtpFT201( 0x18 << 1, cntFT201 );
  sprintf( txt, "MTP word 0x18=%02x\n\r", cntFT201 );
  Serial.print( txt );
}


//======================================================================//
void loop() {
  uint32_t static tim;
  if (  millis() > tim ) {
    tim = millis() + 2000;

    uint16_t static cnt;
    char txt[100];
    sprintf( txt, "Ramka %d\n\r", ++cnt );
    PrintStringFT201( txt );

    sprintf( txt, "Status FT201 %d\n\r",  ReadStFT201() );
    Serial.print( txt );
  }


  //----- Odczytanie liczby znaków w FIFO -----//
  byte RxFIFO = AvailableFT201();
  if ( RxFIFO ) {
#ifdef DEBUG
    char txt[100];
    sprintf( txt, "Odebrano %d\n\r", RxFIFO );
    Serial.print( txt );
#endif
    for (byte x = 0; x < RxFIFO; x++) {
      char c = ReadDataFT201();
      Serial.print( c );
    }
  }


}
